@@ -46,7 +46,6 @@
 #define OPT_VERBOSE		(0x00000001)
 #define OPT_ALL			(0x00000002)
 #define OPT_BY_PID		(0x00000004)
-#define OPT_TOUCH_WRITEABLE	(0x00000008)
 
 #define GOT_MEMFREE		(0x01)
 #define GOT_SWAPFREE		(0x02)
@@ -155,8 +154,9 @@ static int pagein_proc(
 	const int32_t page_size,
 	const pid_t pid,
 	int32_t *const procs,
-	int32_t *const kthreads,
-	int64_t *const total_pages_touched)
+	int64_t *const total_pages_touched,
+	int64_t *const total_pages_skipped,
+	int64_t *const total_mappings)
 {
 	char path[PATH_MAX];
 	char buffer[4096];
@@ -202,7 +202,12 @@ static int pagein_proc(
 		if (sscanf(buffer, "%" SCNx64 "-%" SCNx64
 			   " %5s %*x %*x:%*x %*d %1023s", &begin, &end, prot, tmppath) != 4)
 			continue;
+		(*total_mappings)++;
 		len = end - begin;
+		if (prot[0] != 'r' && prot[3] != 'p') {
+			*total_pages_skipped += (len / page_size);
+			continue;
+		}
 
 		if ((begin >= end) || (len == 0))
 			continue;
@@ -211,8 +216,17 @@ static int pagein_proc(
 
 		for (off = begin; off < end; off += page_size, pages++) {
 			unsigned long data;
+			off_t pos;
 
 			(void)ptrace(PTRACE_PEEKDATA, pid, (void *)off, &data);
+			pos = lseek(fdmem, (off_t)off, SEEK_SET);
+			if (pos == (off_t)off) {
+				size_t sz;
+
+				sz = read(fdmem, &data, sizeof data);
+				(void)sz;
+			}
+
 			pages_touched++;
 		}
 	}
@@ -223,14 +237,12 @@ static int pagein_proc(
 	 *  count these in the stats
 	 */
 	if (has_maps) {
-		*procs += 1;
+		(*procs)++;
 		*total_pages_touched += pages_touched;
-	} else {
-		*kthreads += 1;
 	}
 
 	if (opt_flags & OPT_VERBOSE) {
-		(void)printf("PID:%5d, %12zu pages, %12" PRId64 " pages touched\r", pid, pages,
+		(void)printf("PID:%6d, %12zu pages, %12" PRId64 " pages touched\r", pid, pages,
 			*total_pages_touched);
 		(void)fflush(stdout);
 	}
@@ -250,9 +262,10 @@ err_ret:
 static inline int pagein_all_procs(
 	const int32_t page_size,
 	int32_t *const procs,
-	int32_t *const kthreads,
 	int32_t *const total_procs,
-	int64_t *const total_pages_touched)
+	int64_t *const total_pages_touched,
+	int64_t *const total_pages_skipped,
+	int64_t *const total_mappings)
 {
 	DIR *dp;
 	struct dirent *d;
@@ -267,8 +280,9 @@ static inline int pagein_all_procs(
 		if (isdigit(d->d_name[0]) &&
                     sscanf(d->d_name, "%d", &pid) == 1) {
 			*total_procs += 1;
-			pagein_proc(page_size, pid, procs, kthreads,
-				total_pages_touched);
+			pagein_proc(page_size, pid, procs,
+				total_pages_touched, total_pages_skipped,
+				total_mappings);
 		}
 	}
 
@@ -283,8 +297,10 @@ int main(int argc, char **argv)
 	int64_t swapfree_begin, swapfree_end;
 	int64_t delta;
 	int64_t total_pages_touched = 0ULL;
+	int64_t total_pages_skipped = 0ULL;
+	int64_t total_mappings = 0ULL;
 	static int32_t total_procs = 0;		/* static noclobber */
-	int32_t procs = 0, kthreads = 0;
+	int32_t procs = 0;
 	const int32_t page_size = get_page_size();
 	const int32_t scale = page_size / 1024;
 	static pid_t pid = -1;			/* static noclobber */
@@ -294,7 +310,7 @@ int main(int argc, char **argv)
 	uint8_t	stack[SIGSTKSZ + STACK_ALIGNMENT];
 
 	for (;;) {
-		int c = getopt(argc, argv, "adhp:tvw");
+		int c = getopt(argc, argv, "ahp:v");
 		if (c == -1)
 			break;
 
@@ -317,9 +333,6 @@ int main(int argc, char **argv)
 		case 'v':
 			opt_flags |= OPT_VERBOSE;
 			break;
-		case 'w':
-			opt_flags |= OPT_TOUCH_WRITEABLE;
-			break;
 		default:
 			show_help();
 			exit(EXIT_FAILURE);
@@ -360,13 +373,16 @@ int main(int argc, char **argv)
 		
 	get_memstats(&memfree_begin, &swapfree_begin);
 	if (opt_flags & OPT_ALL)
-		pagein_all_procs(page_size, &procs, &kthreads, &total_procs, &total_pages_touched);
+		pagein_all_procs(page_size, &procs,
+				 &total_procs, &total_pages_touched,
+				 &total_pages_skipped, &total_mappings);
 
 	if (opt_flags & OPT_BY_PID) {
 		int ret;
 
-		ret = pagein_proc(page_size, pid, &procs, &kthreads,
-			&total_pages_touched);
+		ret = pagein_proc(page_size, pid, &procs,
+			&total_pages_touched, &total_pages_skipped,
+			&total_mappings);
 		if (ret < 0) {
 			(void)fprintf(stderr, "cannot page in PID %d errno = %d (%s)\n",
 				pid, -ret, strerror(-ret));
@@ -384,10 +400,11 @@ int main(int argc, char **argv)
 finish:
 	if (opt_flags & OPT_ALL) {
 		(void)printf("Processes scanned:     %" PRIu32 "\n", total_procs);
-		(void)printf("Kernel threads:        %" PRIu32 " (skipped)\n", kthreads);
 		(void)printf("Processes touched:     %" PRIu32 "\n", procs);
 	}
+	(void)printf("Page mappings visited: %" PRIu64 "\n", total_mappings);
 	(void)printf("Pages touched:         %" PRIu64 "\n", total_pages_touched);
+	(void)printf("Pages skipped:         %" PRIu64 " (unreadable)\n", total_pages_skipped);
 	delta = memfree_begin - memfree_end;
 	(void)printf("Free memory decrease:  %" PRId64 "K (%" PRId64 " pages)\n",
 		delta, delta / scale);
